'    QBASIC program for Smart Sensor data transfer demonstration
' ===========================================================================
'
'This example program shows how a communication module can be made.
'It can be used to include more commands or to code in higher
'level languages like in PASCAL or C.
'
'The main loop is similar to EXAMPLE1. Initializating, reading device ID
'and setting emissivity are all done separately. This phase can also include
'other readings (such as read sensor output information, ...) and scans of
'other polling addresses. Also, the temperature reading can be made via
'other commands (such as read current and four dynamic variables, read all
'dynamic variables,...) to get information other than the object temperature.
'
'The COM-port should be selected via the constant COMno before starting.
'Setting the constant, PrintFrames%, to 1 will allow you to monitor the
'bytes passing the HART bus.
'
'To change the emissivity press the "E" key and enter a new value.
'To stop the program press the SPACEBAR or ESC key.


' ----------------------- Internally used contants --------------------------
CONST COMno = 1                                 'Number of serial COM-port
CONST PrintFrames% = 0                          'if > 0 - monitor frames

CONST SlaveTimeOut = &H100, FrameError = &H200
CONST CheckSumError = &H300, ResponseError = &H400

' --------------------- Sensor specific parameter set -----------------------
TYPE SensorRecord                        'Structure for one Sensor
  MfrID AS INTEGER                       'Manufacturer Identifier
  DevType AS INTEGER                     'Device Type Code
  DevID AS LONG                          'Device ID Number
  PreambNo AS INTEGER                    'Number of requested Preambles
  UnivCmdRev AS INTEGER                  'Universal Command Revision
  TrSpecRev AS INTEGER                   'Transmitter Specific Command Rev.
  SoftwareRev AS INTEGER                 'Software Revision
  HardwareRev AS INTEGER                 'Hardware Revision
  PV AS SINGLE                           'Primary Value (Object Temperature)
  PVunit AS INTEGER                      'PV-Unitcode  (C/F)
  Emissivity AS SINGLE                   'Emissivity value of object
END TYPE

' ------------------ Global variable definitions ----------------------------
DIM SHARED Sensor(16) AS SensorRecord    'Sensor Parameter Set Array
COMMON SHARED PollAddr AS INTEGER        'Polling Address used as Index

DIM SHARED TxBuffer(100) AS INTEGER      'Transmit Frame Buffer (Bytes)
DIM SHARED RxBuffer(100) AS INTEGER      'Receive Frame Buffer  (Bytes)
COMMON SHARED TxCount AS INTEGER         'Count of valid Bytes in TxBuffer
COMMON SHARED RxCount AS INTEGER         'Count of valid Bytes in RxBuffer
COMMON SHARED DataPos AS INTEGER         'Position of first Data in RxBuffer

COMMON SHARED Time AS LONG               'BIOS Tick count used for GetTime
COMMON SHARED COMaddr AS INTEGER         'IO Address of used COM-port
COMMON SHARED TransferError AS INTEGER   'indicates communication errors
COMMON SHARED DeviceError AS INTEGER     'first Response Byte of Sensor
COMMON SHARED DeviceStatus AS INTEGER    'second Response Byte of Sensor
COMMON SHARED I AS INTEGER               'standard index variable

' ----------- Declaration of basic subroutines and functions ---------------
DECLARE SUB COMinit ()                   'Initialization of COM-port
DECLARE FUNCTION GetTime& ()             'Get current BIOS tick count
DECLARE FUNCTION DCD% ()                 'Get Data Carrier Detect status
DECLARE SUB SendByte (b AS INTEGER)      'Send one Byte to COM-port
DECLARE SUB SendFrame ()                 'Complete Frame and send it out
DECLARE FUNCTION ReceiveByte% ()         'Get one Byte from COM-port
DECLARE SUB ReceiveFrame ()              'Read in the answer frame
DECLARE SUB PrepareTxBuffer (Cmd AS INTEGER, Cnt AS INTEGER) 'Fill first bytes
DECLARE SUB Transfer ()                  'Handle Request and Answer
DECLARE SUB PrintErrors ()               'Check and print communication errors
DECLARE FUNCTION ReadLong& (p AS INTEGER)   'Build Long from RxBuffer (3Bytes)
DECLARE FUNCTION ReadSingle! (p AS INTEGER) 'Build Single from RxBuffer
DECLARE SUB WriteSingle (S AS SINGLE)       'Place Single Value in TxBuffer
DECLARE FUNCTION UnitStr$ (UnitCode AS INTEGER) 'Convert UnitCode into string

' -------------- Declaration of HART-Command-Routines -----------------------
DECLARE SUB ReadID ()                    'Read Unique Identifier
DECLARE SUB ReadPV ()                    'Read Primary Value(Object Temperature)
DECLARE SUB EEPROMcontrol (Mode AS INTEGER) 'Store all Parameter into EEPROM
DECLARE SUB ReadEmissivity ()            'Read Emissivity value
DECLARE SUB WriteEmissivity (E AS SINGLE) 'Write new Emissivity value


' ***************************** Main program ********************************
CLS
PRINT "Communication Example"
PRINT "====================="
PRINT
COMinit
IF COMaddr <> 0 THEN
  PRINT "COM"; COMno; " initialized !"
ELSE
  PRINT "COM"; COMno; " not available! ": END
END IF

PollAddr = 0                          'Set general access to this device
ReadID                                'Read sensor identifier
PRINT "Sensor No."; Sensor(PollAddr).DevID

ReadEmissivity                      'Read and display emissivity
PRINT USING "Emissivity: #.##"; Sensor(PollAddr).Emissivity

DO                                'Main loop for reading the object temp.
  ReadPV                          'read the object temperature
  PRINT USING "####.#"; Sensor(PollAddr).PV; 'display the object temperature
  PRINT UnitStr(Sensor(PollAddr).PVunit)
  In$ = INKEY$
  IF (In$ = "e") OR (In$ = "E") THEN    'New emissivity entry with "E" key
    PRINT "Old Emissivity=";
    PRINT USING "#.##"; Sensor(PollAddr).Emissivity
    INPUT "New Emissivity=", x
    WriteEmissivity (x)
    REM EEPROMcontrol (0)          'Write parameter into EEPROM (optional)
  END IF
LOOP WHILE (In$ <> " ") AND (In$ <> CHR$(&H1B))


'************************* Subroutines and Functions ************************

' ---------------------------------------------------------------------------
SUB COMinit                                'Initialization of COM-port
  IF (COMno < 1) OR (COMno > 4) THEN : EXIT SUB
  DEF SEG = 0
  COMmemAddr = &H400 + (COMno - 1) * 2     'BIOS listed entry of COM
  COMaddr = PEEK(COMmemAddr) + 256 * PEEK(COMmemAddr + 1)
  IF COMaddr <> 0 THEN
    OUT COMaddr + 3, &H80                  'DLAB=1 -> select divider
    OUT COMaddr, 96                        'Divider = 115200 / 1200 Baud
    OUT COMaddr + 1, 0                     ' - " -
    OUT COMaddr + 3, &HB                   '8 Bit,1 Stopbit,odd Parity,DLAB=0
    OUT COMaddr + 1, 0                     'Disable all interrupts
    OUT COMaddr + 4, 1                     'Switch DTR to positive
  END IF
END SUB

' ---------------------------------------------------------------------------
FUNCTION DCD%                           'Get carrier status
  DCD = (INP(COMaddr + 6) AND &H80)     'DCD<>0 -> carrier active
END FUNCTION

' ---------------------------------------------------------------------------
SUB EEPROMcontrol (Mode AS INTEGER)     'EEPROM parameter transfer
  PrepareTxBuffer 39, 1                 'Cmd: EEPROMcontrol, Cnt:1
  TxBuffer(TxCount) = Mode              '0:EEPROMwrite, 1:EEPROMread
  TxCount = TxCount + 1
  Transfer
END SUB

' ---------------------------------------------------------------------------
FUNCTION GetTime&                           'Get 55 ms time counter of BIOS
  DEF SEG = 0
  Time = PEEK(&H46F) * &H100 + PEEK(&H46E)
  Time = (Time * &H100 + PEEK(&H46D)) * &H100 + PEEK(&H46C)
  GetTime = Time
END FUNCTION

' ---------------------------------------------------------------------------
SUB PrepareTxBuffer (Cmd AS INTEGER, Cnt AS INTEGER)
  TxBuffer(0) = &H82                             'long address request
  TxBuffer(1) = Sensor(PollAddr).MfrID AND &H3F  'insert long address
  TxBuffer(2) = Sensor(PollAddr).DevType                         ' -"-
  TxBuffer(3) = (Sensor(PollAddr).DevID AND &HFF0000) / &H10000  ' -"-
  TxBuffer(4) = (Sensor(PollAddr).DevID AND &HFF00) / &H100      ' -"-
  TxBuffer(5) = Sensor(PollAddr).DevID AND &HFF                  ' -"-
  TxBuffer(6) = Cmd                     'Command No
  TxBuffer(7) = Cnt                     'Byte Count
  TxCount = 8
END SUB

' ---------------------------------------------------------------------------
SUB PrintErrors                 'Display Errors and Warnings if occured
  IF TransferError <> 0 THEN
    PRINT "Communication Error: ";
    IF (TransferError AND &HF) <> 0 THEN
      PRINT "UART Receive Error"
    ELSEIF TransferError = SlaveTimeOut THEN
      PRINT "Slave Timeout"
    ELSEIF TransferError = FrameError THEN
      PRINT "Frame Error"
    ELSEIF TransferError = CheckSumError THEN
      PRINT "Checksum Error"
    ELSEIF TransferError = ResponseError THEN
      PRINT "Device Error Code: "; HEX$(DeviceError)
    END IF

  ELSEIF DeviceError <> 0 THEN
    PRINT "Device Warning Code:"; DeviceError;
    IF DeviceError = 2 THEN
      PRINT ": Invalid Selection"
    ELSEIF DeviceError = 3 THEN
      PRINT ": Passed Parameter is too large"
    ELSEIF DeviceError = 4 THEN
      PRINT ": Passed parameter is too small"
    ELSEIF DeviceError = 5 THEN
      PRINT ": Too few data bytes received"
    ELSEIF DeviceError = 7 THEN
      PRINT ": Device in write-protected mode"
    ELSEIF DeviceError = 8 THEN
      PRINT ": Set to nearest possible value"
    ELSEIF DeviceError = 9 THEN
      PRINT ": Lower range value too high"
    ELSEIF DeviceError = 10 THEN
      PRINT ": Lower range value too low"
    ELSEIF DeviceError = 11 THEN
      PRINT ": Upper range value too high"
    ELSEIF DeviceError = 12 THEN
      PRINT ": Upper range value too low"
    ELSEIF DeviceError = 14 THEN
      PRINT ": Span too small"
    ELSEIF DeviceError = 32 THEN
      PRINT ": Device busy"
    ELSEIF DeviceError = 64 THEN
      PRINT ": Command not implemented"
    PRINT
    END IF
  END IF
END SUB

' ---------------------------------------------------------------------------
SUB ReadEmissivity
  PrepareTxBuffer 128, 0                 'Cmd:Read Emissivity, Cnt:0
  Transfer
  IF TransferError = 0 THEN
    Sensor(PollAddr).Emissivity = ReadSingle(DataPos)
  END IF
END SUB

' ---------------------------------------------------------------------------
SUB ReadID
  TxBuffer(0) = 2                                'Request short format
  TxBuffer(1) = PollAddr                         'Polling Address
  TxBuffer(2) = 0                                'Command 0: Read ID
  TxBuffer(3) = 0                                'Byte Count = 0
  TxCount = 4
  Transfer
  IF TransferError <> 0 THEN : EXIT SUB
  Sensor(PollAddr).MfrID = RxBuffer(DataPos + 1)       'Manufacturer ID
  Sensor(PollAddr).DevType = RxBuffer(DataPos + 2)     'Device Type
  Sensor(PollAddr).PreambNo = RxBuffer(DataPos + 3)    'Number of Preambles
  Sensor(PollAddr).UnivCmdRev = RxBuffer(DataPos + 4)  'Universal Command Rev.
  Sensor(PollAddr).TrSpecRev = RxBuffer(DataPos + 5)   'Trans.Specific Cmd.Rev.
  Sensor(PollAddr).SoftwareRev = RxBuffer(DataPos + 6) 'Software Revision
  Sensor(PollAddr).HardwareRev = RxBuffer(DataPos + 7) 'Hardware Revision
  Sensor(PollAddr).DevID = ReadLong&(DataPos + 9)      'Device ID number
END SUB

' ---------------------------------------------------------------------------
FUNCTION ReadLong& (p AS INTEGER)   'Read 3 byte long integer from Rx Buffer
'
  l& = RxBuffer(p)
  l& = l& * &H100 + RxBuffer(p + 1)
  l& = l& * &H100 + RxBuffer(p + 2)
  ReadLong = l&
END FUNCTION

' ---------------------------------------------------------------------------
SUB ReadPV                              'Read Object Temperature
  PrepareTxBuffer 1, 0                  'Cmd: ReadPV, Cnt:0
  Transfer
  IF TransferError <> 0 THEN : EXIT SUB
  Sensor(PollAddr).PVunit = RxBuffer(DataPos)   'Unit code Celsius/Fahrenheit
  Sensor(PollAddr).PV = ReadSingle(DataPos + 1) 'Get value
END SUB

' ---------------------------------------------------------------------------
FUNCTION ReadSingle! (p AS INTEGER) 'Read floating point value from Rx Buffer
  DIM buffer AS STRING * 4
  MID$(buffer, 1, 1) = CHR$(RxBuffer(p + 3))  'Mirror rank of bytes
  MID$(buffer, 2, 1) = CHR$(RxBuffer(p + 2))
  MID$(buffer, 3, 1) = CHR$(RxBuffer(p + 1))
  MID$(buffer, 4, 1) = CHR$(RxBuffer(p))
  ReadSingle = CVS(buffer)                    'Convert to Single
END FUNCTION

' ---------------------------------------------------------------------------
FUNCTION ReceiveByte%                 'Read byte from COM-port
  ByteTime = GetTime
  DO                                  'Wait until received or carrier is off
    dt = GetTime - ByteTime           '  or time out
  LOOP WHILE ((INP(COMaddr + 5) AND 1) = 0) AND (DCD <> 0) AND (dt < 2)

  ReceiveByte = INP(COMaddr)
  IF PrintFrames% > 0 THEN : PRINT HEX$(INP(COMaddr)); ","; 'For monitoring
  IF dt >= 2 THEN : TransferError = SlaveTimeOut

  'Include UART error flags
  TransferError = TransferError OR (INP(COMaddr + 4) AND &H1E)
END FUNCTION

' ---------------------------------------------------------------------------
SUB ReceiveFrame                            'Read and check answer frame
  DEFINT B
  AnswerTime = GetTime
  WHILE (DCD = 0) AND (TransferError = 0)   'Wait until sensor carrier on
    IF (GetTime - AnswerTime) >= 5 THEN     'TimeOut is about 5*55ms
      TransferError = SlaveTimeOut
      EXIT SUB
    END IF
  WEND

  DO
    DO
      b = ReceiveByte
      TransferError = TransferError AND &HFFFFFFF0 'Suppress UART errors
      IF DCD = 0 THEN : TransferError = SlaveTimeOut
      IF TransferError <> 0 THEN : EXIT SUB
    LOOP WHILE b <> &HFF                    'Read until valid preamble
    DO
      b = ReceiveByte
      IF DCD = 0 THEN : TransferError = SlaveTimeOut
      IF TransferError <> 0 THEN : EXIT SUB
    LOOP WHILE b = &HFF                     'Read until no Preamble
  LOOP WHILE (b <> &H6) AND (b <> &H86)     'Repeat until valid start char

  RxBuffer(0) = b                           'Store start haracter
  RxCount = 1
  WHILE DCD <> 0                            'Get all datas of the frame
    b = ReceiveByte
    RxBuffer(RxCount) = b                   'Store answer bytes in buffer
    IF TransferError <> 0 THEN : EXIT SUB
    RxCount = RxCount + 1
  WEND

  IF RxBuffer(0) = &H6 THEN                 'Short address format ?
    ByteCountPos = 3
  ELSEIF RxBuffer(0) = &H86 THEN            'Long address format ?
    ByteCountPos = 7
  END IF

  FrameEndPos = RxBuffer(ByteCountPos) + ByteCountPos + 1
  IF RxCount < FrameEndPos THEN             'Received too few data bytes ?
    TransferError = FrameError
    EXIT SUB
  END IF

  CheckByte = 0
  FOR I = 0 TO FrameEndPos                  'Compare checksum
    CheckByte = CheckByte XOR RxBuffer(I)
  NEXT
  IF CheckByte <> 0 THEN
    TransferError = CheckSumError           'Mark error
    EXIT SUB
  END IF

  DeviceError = RxBuffer(ByteCountPos + 1)  'Response Status Bytes
  DeviceStatus = RxBuffer(ByteCountPos + 2)
  DataPos = ByteCountPos + 3                'Index to data

  IF DeviceError > &H80 THEN                'Did sensor get correct frame?
    TransferError = ResponseError
  END IF
END SUB

' ---------------------------------------------------------------------------
SUB SendByte (b AS INTEGER)              'Send out the byte to COM-port
  WHILE (INP(COMaddr + 5) AND &H40) = 0
  WEND                                   'Wait until transmit buffer empty
  OUT COMaddr, b
  IF PrintFrames% > 0 THEN : PRINT HEX$(b AND &HFF); ",";  'For monitoring
END SUB

' ---------------------------------------------------------------------------
SUB SendFrame                                'Send out request frame
  BusFreeTime& = Time                        'Time count since last transfer
  DO
    IF DCD <> 0 THEN BusFreeTime& = GetTime
  LOOP WHILE (Time - BusFreeTime&) >= 2      'Until bus is free

  OUT COMaddr + 4, 3                         'Switch PC's carrier on
  FOR I = 1 TO 3
    SendByte (&HFF)                          'Send preamble
  NEXT

  CheckByte = 0
  FOR I = 0 TO TxCount - 1
    CheckByte = CheckByte XOR TxBuffer(I)    'Calculate checksum
    SendByte (TxBuffer(I))                   'Send all bytes of frame
  NEXT
  SendByte (CheckByte)

  WHILE (INP(COMaddr + 5) AND &H60) <> &H60  'Wait until all bytes transfered
  WEND

  OUT COMaddr + 4, 1                         'Carrier off
  WHILE (DCD <> 0)                           'Wait until reflection of
  WEND                                       '  PC's carrier is gone
END SUB

' ---------------------------------------------------------------------------
SUB Transfer                     'Send request frames and read in response
  Try = 0
  DO                             'Try for maximum of 3 times
    TransferError = 0            'Reset Transfer Error Code
    SendFrame
    IF PrintFrames% > 0 THEN : PRINT  'Only to monitor
    ReceiveFrame
    IF PrintFrames% > 0 THEN : PRINT  'Only to monitor
    Try = Try + 1
  LOOP WHILE (Try <= 3) AND (TransferError <> 0)
  PrintErrors                      'If needed, Display errors and warnings
 
  IF TransferError <> 0 THEN : END    'Leave program
  IF PrintFrames% > 0 THEN
    PRINT "Press any key to continue !":
    DO: In$ = INKEY$: LOOP WHILE In$ = ""
    IF In$ = CHR$(&H1B) THEN END
  END IF
END SUB

' ---------------------------------------------------------------------------
FUNCTION UnitStr$ (UnitCode AS INTEGER)  'Decoding HART Unit Codes
  IF UnitCode = 32 THEN
    UnitStr = "C"
  ELSEIF UnitCode = 33 THEN
    UnitStr = "F"
  ELSE UnitStr = "?"                      'Other devices
  END IF
END FUNCTION

' ---------------------------------------------------------------------------
SUB WriteEmissivity (S AS SINGLE)       'Set emissivity
  PrepareTxBuffer 129, 4                'Cmd: Write Emissivity, Cnt:4
  WriteSingle (S)                       'Append new value to buffer
  Transfer
  IF (TransferError OR DeviceError) <> 0 THEN : EXIT SUB
  Sensor(PollAddr).Emissivity = ReadSingle(DataPos)  'Get value
END SUB

' ---------------------------------------------------------------------------
SUB WriteSingle (S AS SINGLE)        'Append floating point number to buffer
  DIM buffer AS STRING * 4
  buffer = MKS$(S)                                 'Convert into string
  TxBuffer(TxCount) = ASC(MID$(buffer, 4, 1))      'Mirror rank of bytes
  TxBuffer(TxCount + 1) = ASC(MID$(buffer, 3, 1))
  TxBuffer(TxCount + 2) = ASC(MID$(buffer, 2, 1))
  TxBuffer(TxCount + 3) = ASC(MID$(buffer, 1, 1))
  TxCount = TxCount + 4
END SUB

